from enum import Enum
from warnings import warn

from _pytest.config import Config


class Scope(Enum):
    CLASS = 1
    MODULE = 2
    SESSION = 3


class Settings:
    """Holds all configuration settings."""

    valid_scopes = {
        "class": Scope.CLASS,
        "module": Scope.MODULE,
        "session": Scope.SESSION,
    }

    def __init__(self, config: Config) -> None:
        self.sparse_ordering: bool = config.getoption("sparse_ordering")
        self.order_dependencies: bool = config.getoption("order_dependencies")
        self.marker_prefix: str = config.getoption("order_marker_prefix")
        self.error_on_failed_ordering: str = config.getoption(
            "error_on_failed_ordering"
        )
        scope: str = config.getoption("order_scope")
        if scope in self.valid_scopes:
            self.scope: Scope = self.valid_scopes[scope]
        else:
            if scope is not None:
                warn(
                    f"Unknown order scope '{scope}', ignoring it. "
                    "Valid scopes are 'session', 'module' and 'class'."
                )
            self.scope = Scope.SESSION
        scope_level: int = config.getoption("order_scope_level") or 0
        if scope_level != 0 and self.scope != Scope.SESSION:
            warn(
                f"order-scope-level cannot be used together with --order-scope={scope}"
            )
            scope_level = 0
        self.scope_level: int = scope_level
        group_scope: str = config.getoption("order_group_scope")
        if group_scope in self.valid_scopes:
            self.group_scope: Scope = self.valid_scopes[group_scope]
        else:
            if group_scope is not None:
                warn(
                    f"Unknown order group scope '{group_scope}', ignoring it. "
                    "Valid scopes are 'session', 'module' and 'class'."
                )
            self.group_scope = self.scope
        if self.group_scope.value > self.scope.value:
            warn("Group scope is larger than order scope, ignoring it.")
            self.group_scope = self.scope
        try:
            auto_mark_dep = config.getini("automark_dependency")
            if isinstance(auto_mark_dep, str):
                auto_mark_dep = auto_mark_dep.lower() in (
                    "1",
                    "yes",
                    "y",
                    "true",
                    "t",
                    "on",
                )
        except ValueError:
            auto_mark_dep = False
        self.auto_mark_dep = auto_mark_dep



import sys
from typing import Optional, List, Dict, Tuple, Generic, TypeVar

from _pytest.python import Function

from .settings import Scope, Settings


_ItemType = TypeVar("_ItemType", "Item", "ItemGroup")


class Item:
    """Represents a single test item."""

    def __init__(self, item: Function) -> None:
        self.item: Function = item
        self.nr_rel_items: int = 0
        self.order: Optional[int] = None
        self._node_id: Optional[str] = None

    def inc_rel_marks(self) -> None:
        if self.order is None:
            self.nr_rel_items += 1

    def dec_rel_marks(self) -> None:
        if self.order is None:
            self.nr_rel_items -= 1

    @property
    def module_path(self) -> str:
        return self.item.nodeid[: self.node_id.index("::")]

    def parent_path(self, level) -> str:
        return "/".join(self.module_path.split("/")[:level])

    @property
    def node_id(self) -> str:
        if self._node_id is None:
            # in pytest < 4 the nodeid has an unwanted ::() part
            self._node_id = self.item.nodeid.replace("::()", "")
        return self._node_id


class ItemList:
    """Handles a group of items with the same scope."""

    def __init__(
        self,
        items: List[Item],
        settings: Settings,
        scope: Scope,
        rel_marks: List["RelativeMark[Item]"],
        dep_marks: List["RelativeMark[Item]"],
    ) -> None:
        self.items = items
        self.settings = settings
        self.scope = scope
        self.start_items: List[Tuple[int, List[Item]]] = []
        self.end_items: List[Tuple[int, List[Item]]] = []
        self.unordered_items: List[Item] = []
        self._start_items: Dict[int, List[Item]] = {}
        self._end_items: Dict[int, List[Item]] = {}
        self.all_rel_marks = rel_marks
        self.all_dep_marks = dep_marks
        self.rel_marks = filter_marks(rel_marks, items)
        self.dep_marks = filter_marks(dep_marks, items)

    def collect_markers(self, item: Item) -> None:
        self.handle_order_mark(item)
        if item.nr_rel_items or item.order is None:
            self.unordered_items.append(item)

    def handle_order_mark(self, item: Item) -> None:
        if item.order is not None:
            if item.order < 0:
                self._end_items.setdefault(item.order, []).append(item)
            else:
                self._start_items.setdefault(item.order, []).append(item)

    def sort_numbered_items(self) -> List[Item]:
        self.start_items = sorted(self._start_items.items())
        self.end_items = sorted(self._end_items.items())
        sorted_list = []
        index = 0
        for order, items in self.start_items:
            if self.settings.sparse_ordering:
                while order > index and self.unordered_items:
                    sorted_list.append(self.unordered_items.pop(0))
                    index += 1
            sorted_list += items
            index += len(items)
        mid_index = len(sorted_list)
        index = -1
        for order, items in reversed(self.end_items):
            if self.settings.sparse_ordering:
                while order < index and self.unordered_items:
                    sorted_list.insert(mid_index, self.unordered_items.pop())
                    index -= 1
            sorted_list[mid_index:mid_index] = items
            index -= len(items)
        sorted_list[mid_index:mid_index] = self.unordered_items
        return sorted_list

    def print_unhandled_items(self) -> None:
        failed_items = [mark.item for mark in self.rel_marks] + [
            mark.item for mark in self.dep_marks
        ]
        msg = " ".join([item.node_id for item in failed_items])
        sys.stdout.write("\nWARNING: cannot execute test relative to others: ")
        sys.stdout.write(msg)
        if self.settings.error_on_failed_ordering:
            sys.stdout.write(" - ignoring the marker.\n")
        else:
            sys.stdout.write(".\n")
        sys.stdout.flush()
        if self.settings.error_on_failed_ordering:
            for item in failed_items:
                item.item.fixturenames.insert(0, "fail_after_cannot_order")

    def number_of_rel_groups(self) -> int:
        return len(self.rel_marks) + len(self.dep_marks)

    def handle_rel_marks(self, sorted_list: List[Item]) -> None:
        self.handle_relative_marks(self.rel_marks, sorted_list, self.all_rel_marks)

    def handle_dep_marks(self, sorted_list: List[Item]) -> None:
        self.handle_relative_marks(self.dep_marks, sorted_list, self.all_dep_marks)

    @staticmethod
    def handle_relative_marks(
        marks: List["RelativeMark[Item]"],
        sorted_list: List[Item],
        all_marks: List["RelativeMark[Item]"],
    ) -> None:
        for mark in reversed(marks):
            if move_item(mark, sorted_list):
                marks.remove(mark)
                all_marks.remove(mark)

    def group_order(self) -> Optional[int]:
        if self.start_items:
            return self.start_items[0][0]
        elif self.end_items:
            return self.end_items[-1][0]
        return None


class ItemGroup:
    """
    Holds a group of sorted items with the same group order scope.
    Used for sorting groups similar to Item for sorting items.
    """

    def __init__(
        self, items: Optional[List[Item]] = None, order: Optional[int] = None
    ) -> None:
        self.items: List[Item] = items or []
        self.order = order
        self.nr_rel_items = 0

    def inc_rel_marks(self) -> None:
        if self.order is None:
            self.nr_rel_items += 1

    def dec_rel_marks(self) -> None:
        if self.order is None:
            self.nr_rel_items -= 1

    def extend(self, groups: List["ItemGroup"], order: Optional[int]) -> None:
        for group in groups:
            self.items.extend(group.items)
        self.order = order


class RelativeMark(Generic[_ItemType]):
    """
    Represents a marker for an item or an item group.
    Holds two related items or groups and their relationship.
    """

    def __init__(
        self,
        item: _ItemType,
        item_to_move: _ItemType,
        move_after: bool,
    ) -> None:
        self.item: _ItemType = item
        self.item_to_move: _ItemType = item_to_move
        self.move_after: bool = move_after


def filter_marks(
    marks: List[RelativeMark[_ItemType]], all_items: List[Item]
) -> List[RelativeMark[_ItemType]]:
    result = []
    for mark in marks:
        if mark.item in all_items and mark.item_to_move in all_items:
            result.append(mark)
        else:
            mark.item_to_move.dec_rel_marks()
    return result


def move_item(mark: RelativeMark[_ItemType], sorted_items: List[_ItemType]) -> bool:
    if (
        mark.item not in sorted_items
        or mark.item_to_move not in sorted_items
        or mark.item.nr_rel_items
    ):
        return False
    pos_item = sorted_items.index(mark.item)
    pos_item_to_move = sorted_items.index(mark.item_to_move)
    if mark.item_to_move.order is not None and mark.item.order is None:
        # if the item to be moved has already been ordered numerically,
        # and the other item is not ordered, we move that one instead
        mark.move_after = not mark.move_after
        mark.item, mark.item_to_move = mark.item_to_move, mark.item
        pos_item, pos_item_to_move = pos_item_to_move, pos_item
    mark.item_to_move.dec_rel_marks()
    if mark.move_after:
        if pos_item_to_move < pos_item + 1:
            del sorted_items[pos_item_to_move]
            sorted_items.insert(pos_item, mark.item_to_move)
    else:
        if pos_item_to_move > pos_item:
            del sorted_items[pos_item_to_move]
            pos_item -= 1
            sorted_items.insert(pos_item + 1, mark.item_to_move)
    return True



from typing import List

import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Parser
from _pytest.main import Session
from _pytest.mark import Mark
from _pytest.python import Function

from .sorter import Sorter


def pytest_configure(config: Config) -> None:
    """
    Register the "order" marker and configure the plugin,
    depending on the CLI options.
    """

    provided_by_pytest_order = (
        "Provided by pytest-order. " "See also: https://pytest-order.readthedocs.io/"
    )

    config_line = (
        "order: specify ordering information for when tests should run "
        "in relation to one another. " + provided_by_pytest_order
    )
    config.addinivalue_line("markers", config_line)
    # We need to dynamically add this `tryfirst` decorator to the plugin:
    # only when the CLI option is present should the decorator be added.
    # Thus, we manually run the decorator on the class function and
    # manually replace it.
    if config.getoption("indulgent_ordering"):
        wrapper = pytest.hookimpl(tryfirst=True)
    else:
        wrapper = pytest.hookimpl(trylast=True)
    OrderingPlugin.pytest_collection_modifyitems = wrapper(  # type:ignore[attr-defined]
        modify_items
    )
    config.pluginmanager.register(OrderingPlugin(), "orderingplugin")


def pytest_addoption(parser: Parser) -> None:
    """Set up CLI option for pytest"""
    group = parser.getgroup("order")
    group.addoption(
        "--indulgent-ordering",
        action="store_true",
        dest="indulgent_ordering",
        help=(
            "Request that the sort order provided by pytest-order be applied "
            "before other sorting, allowing the other sorting to have priority"
        ),
    )
    group.addoption(
        "--order-scope",
        action="store",
        dest="order_scope",
        help=(
            "Defines the scope used for ordering. Possible values are: "
            "'session' (default), 'module', and 'class'. "
            "Ordering is only done inside a scope."
        ),
    )
    group.addoption(
        "--order-scope-level",
        action="store",
        type=int,
        dest="order_scope_level",
        help=(
            "Defines that the given directory level is used as order scope. "
            "Cannot be used with --order-scope. The value is a number "
            "that defines the hierarchical index of the directories used as "
            "order scope, starting with 0 at session scope."
        ),
    )
    group.addoption(
        "--order-group-scope",
        action="store",
        dest="order_group_scope",
        help=(
            "Defines the scope used for order groups. Possible values are: "
            " 'session' (default), 'module', and 'class'. "
            "Ordering is first done inside a group, then between groups."
        ),
    )
    group.addoption(
        "--sparse-ordering",
        action="store_true",
        dest="sparse_ordering",
        help=(
            "If there are gaps between ordinals, they are filled "
            "with unordered tests."
        ),
    )
    group.addoption(
        "--order-dependencies",
        action="store_true",
        dest="order_dependencies",
        help=(
            "If set, dependencies added by pytest-dependency will be ordered "
            "if needed."
        ),
    )
    group.addoption(
        "--order-marker-prefix",
        action="store",
        dest="order_marker_prefix",
        help=(
            "If set, markers starting with the given prefix followed by a number "
            "are handled like order markers with an index."
        ),
    )
    group.addoption(
        "--error-on-failed-ordering",
        action="store_true",
        dest="error_on_failed_ordering",
        help=(
            "If set, tests with relative markers that could not be ordered "
            "will error instead of generating only a warning."
        ),
    )


def _get_mark_description(mark: Mark):
    if mark.kwargs:
        return ", ".join([f"{k}={v}" for k, v in mark.kwargs.items()])
    elif mark.args:
        return f"index={mark.args[0]}"
    return mark


def pytest_generate_tests(metafunc):
    """
    Handle multiple pytest.mark.order decorators.

    Make parametrized tests with corresponding order marks.
    """
    if getattr(metafunc, "function", False):
        if getattr(metafunc.function, "pytestmark", False):
            # Get list of order marks
            marks = metafunc.function.pytestmark
            order_marks = [mark for mark in marks if mark.name == "order"]
            if len(order_marks) > 1:
                # Remove all order marks
                metafunc.function.pytestmark = [
                    mark for mark in marks if mark.name != "order"
                ]
                # Prepare arguments for parametrization with order marks
                args = [
                    pytest.param(_get_mark_description(mark), marks=[mark])
                    for mark in order_marks
                ]
                if "order" not in metafunc.fixturenames:
                    metafunc.fixturenames.append("order")
                metafunc.parametrize("order", args)


@pytest.fixture
def fail_after_cannot_order():
    pytest.fail("The test could not be ordered")


class OrderingPlugin:
    """
    Plugin implementation.

    By putting this in a class, we are able to dynamically register it after
    the CLI is parsed.
    """


def modify_items(session: Session, config: Config, items: List[Function]) -> None:
    sorter = Sorter(config, items)
    items[:] = sorter.sort_items()



import re
import sys
from collections import OrderedDict
from contextlib import suppress
from typing import Optional, List, Dict, Tuple, cast
from warnings import warn

from _pytest.config import Config
from _pytest.mark import Mark
from _pytest.python import Function

from .item import Item, ItemList, ItemGroup, filter_marks, move_item, RelativeMark
from .settings import Settings, Scope

orders_map = {
    "first": 0,
    "second": 1,
    "third": 2,
    "fourth": 3,
    "fifth": 4,
    "sixth": 5,
    "seventh": 6,
    "eighth": 7,
    "last": -1,
    "second_to_last": -2,
    "third_to_last": -3,
    "fourth_to_last": -4,
    "fifth_to_last": -5,
    "sixth_to_last": -6,
    "seventh_to_last": -7,
    "eighth_to_last": -8,
}


class Sorter:
    """
    Sort all items according to the given configuration.
    """

    def __init__(self, config: Config, items: List[Function]) -> None:
        self.settings: Settings = Settings(config)
        self.items: List[Item] = [Item(item) for item in items]
        self.node_ids: Dict[str, Item] = OrderedDict()
        self.node_id_last: Dict[str, List[str]] = {}
        for item in self.items:
            self.node_ids[item.node_id] = item
            last_part = item.node_id.rpartition("::")[2]
            if "[" in last_part:
                last_part = last_part.rpartition("[")[0]
            # save last nodeid component to avoid to iterate over all
            # items for each label
            self.node_id_last.setdefault(last_part, []).append(item.node_id)
        self.rel_marks: List[RelativeMark[Item]] = []
        self.dep_marks: List[RelativeMark[Item]] = []

    def sort_items(self) -> List[Function]:
        """
        Do the actual sorting and return the sorted items.
        """
        self.collect_markers()
        if self.settings.scope == Scope.SESSION:
            if self.settings.scope_level > 0:
                dir_groups = directory_item_groups(
                    self.items, self.settings.scope_level
                )
                sorted_list = []
                for items in dir_groups.values():
                    sorter = ScopeSorter(
                        self.settings, items, self.rel_marks, self.dep_marks
                    )
                    sorted_list.extend(sorter.sort_items())
            else:
                sorter = ScopeSorter(
                    self.settings,
                    self.items,
                    self.rel_marks,
                    self.dep_marks,
                    session_scope=True,
                )
                sorted_list = sorter.sort_items()
        elif self.settings.scope == Scope.MODULE:
            module_groups = module_item_groups(self.items)
            sorted_list = []
            for module_items in module_groups.values():
                sorter = ScopeSorter(
                    self.settings, module_items, self.rel_marks, self.dep_marks
                )
                sorted_list.extend(sorter.sort_items())
        else:  # class scope
            class_groups = class_item_groups(self.items)
            sorted_list = []
            for class_items in class_groups.values():
                sorter = ScopeSorter(
                    self.settings, class_items, self.rel_marks, self.dep_marks
                )
                sorted_list.extend(sorter.sort_items())
        return [item.item for item in sorted_list]

    def mark_binning(
        self,
        item: Item,
        dep_marks: Dict[Tuple[str, Scope, str], List[Item]],
        aliases: Dict[str, List[Item]],
    ) -> None:
        """
        Collect relevant markers for the given item.
        """
        keys = item.item.keywords.keys()
        has_dependency = "dependency" in keys
        has_order = "order" in keys
        if not has_order and self.settings.marker_prefix:
            for key in keys:
                if key.startswith(self.settings.marker_prefix):
                    try:
                        index = int(key[len(self.settings.marker_prefix)])
                        item.order = index
                    except ValueError:
                        pass
        if has_dependency or self.settings.auto_mark_dep:
            self.handle_dependency_mark(item, has_order, dep_marks, aliases)
        if has_order:
            self.handle_order_marks(item)

    def handle_dependency_mark(
        self,
        item: Item,
        has_order: bool,
        dep_marks: Dict[Tuple[str, Scope, str], List[Item]],
        aliases: Dict[str, List[Item]],
    ) -> None:
        # always order dependencies if an order mark is present
        # otherwise only if order-dependencies is set
        mark = item.item.get_closest_marker("dependency")
        name_mark = None
        if mark and (self.settings.order_dependencies or has_order):
            dependent_mark = mark.kwargs.get("depends")
            if dependent_mark:
                scope = scope_from_name(mark.kwargs.get("scope", "module"))
                prefix = scoped_node_id(item.node_id, scope)
                for name in dependent_mark:
                    dep_marks.setdefault((name, scope, prefix), []).append(item)
                    item.inc_rel_marks()
            # we always collect the names of the dependent items, because
            # we need them in both cases
            name_mark = mark.kwargs.get("name")
        # the default name in pytest-dependency is the nodeid or a part
        # of the nodeid, depending on the scope
        if not name_mark:
            name_mark = item.node_id
        aliases.setdefault(name_mark, []).append(item)

    def handle_order_marks(self, item: Item) -> None:
        marks = item.item.iter_markers("order")
        for mark in marks:
            self.handle_order_mark(item, mark)

    def handle_order_mark(self, item: Item, mark: Mark) -> None:
        order = mark.args[0] if mark.args else mark.kwargs.get("index")
        if order is not None:
            if isinstance(order, int):
                order = int(order)
            elif order in orders_map:
                order = orders_map[order]
            else:
                warn(f"Unknown order attribute:'{order}'")
                order = None
        if item.order is None:
            item.order = order
        self.handle_relative_marks(item, mark)
        if order is not None:
            item.nr_rel_items = 0

    def items_from_label(self, label: str, item: Item, is_cls_mark: bool) -> List[Item]:
        """
        Return the list of matching items from the given label.
        The list contains one item for a single matching test, several items
        in the case of a matching parametrized test, or no item in case of
        an invalid label.
        """
        item_id = item.node_id
        label_len = len(label)
        last_comp = label.split("/")[-1].split("::")[-1]
        items = []
        with suppress(KeyError):
            node_ids = self.node_id_last[last_comp]
            for node_id in node_ids:
                if node_id.endswith(label):
                    id_start = node_id[:-label_len]
                elif node_id.endswith("]") and node_id.rpartition("[")[0].endswith(
                    label
                ):
                    id_start = node_id.rpartition("[")[0][:-label_len]
                else:
                    continue
                if is_cls_mark and id_start.count("::") == 2:
                    continue
                if item_id.startswith(id_start):
                    items.append(self.node_ids[node_id])

        return items

    def items_from_class_label(self, label: str, item: Item) -> List[Item]:
        items = []
        item_id = item.node_id
        label_len = len(label)
        for node_id in self.node_ids:
            if node_id.count("::") == 2:
                cls_index = node_id.rindex("::")
                if node_id[:cls_index].endswith(label):
                    id_start = node_id[: cls_index - label_len]
                    if item_id.startswith(id_start):
                        items.append(self.node_ids[node_id])
        return items

    def handle_before_or_after_mark(
        self, item: Item, mark: Mark, marker_name: str, is_after: bool
    ) -> bool:
        def is_class_mark() -> bool:
            if item.item.cls and item.item.parent:
                return item.item.parent.get_closest_marker("order") == mark
            return False

        def is_mark_for_class() -> bool:
            return "::" not in marker_name and is_class_mark()

        is_cls_mark = is_class_mark()
        items_for_label = self.items_from_label(marker_name, item, is_cls_mark)
        if items_for_label:
            for item_for_label in items_for_label:
                rel_mark = RelativeMark(item_for_label, item, move_after=is_after)
                if is_after or not is_cls_mark:
                    self.rel_marks.append(rel_mark)
                else:
                    self.rel_marks.insert(0, rel_mark)
                item.inc_rel_marks()
            return True
        else:
            if is_mark_for_class():
                items = self.items_from_class_label(marker_name, item)
                for item_for_label in items:
                    rel_mark = RelativeMark(item_for_label, item, move_after=is_after)
                    if is_after:
                        self.rel_marks.append(rel_mark)
                    else:
                        self.rel_marks.insert(0, rel_mark)
                    item.inc_rel_marks()
                return len(items) > 0
        return False

    def handle_relative_marks(self, item: Item, mark: Mark) -> bool:
        has_relative_marks = False
        before_marks = mark.kwargs.get("before", ())
        if before_marks and not isinstance(before_marks, (list, tuple)):
            before_marks = (before_marks,)
        for before_mark in before_marks:
            if self.handle_before_or_after_mark(
                item, mark, before_mark, is_after=False
            ):
                has_relative_marks = True
            else:
                self.warn_about_unknown_test(item, before_mark)
        after_marks = mark.kwargs.get("after", ())
        if after_marks and not isinstance(after_marks, (list, tuple)):
            after_marks = (after_marks,)
        for after_mark in after_marks:
            if self.handle_before_or_after_mark(item, mark, after_mark, is_after=True):
                has_relative_marks = True
            else:
                self.warn_about_unknown_test(item, after_mark)
        return has_relative_marks

    def warn_about_unknown_test(self, item: Item, rel_mark: str) -> None:
        if self.settings.error_on_failed_ordering:
            item.item.fixturenames.insert(0, "fail_after_cannot_order")
            ignore_msg = ""
        else:
            ignore_msg = " - ignoring the marker"
        sys.stdout.write(
            f"\nWARNING: cannot execute '{item.item.name}' relative to others: "
            f"'{rel_mark}'{ignore_msg}."
        )

    def collect_markers(self) -> None:
        aliases: Dict[str, List[Item]] = {}
        dep_marks: Dict[Tuple[str, Scope, str], List[Item]] = {}
        for item in self.items:
            self.mark_binning(item, dep_marks, aliases)
        self.resolve_dependency_markers(dep_marks, aliases)

    def resolve_dependency_markers(
        self,
        dep_marks: Dict[Tuple[str, Scope, str], List[Item]],
        aliases: Dict[str, List[Item]],
    ) -> None:
        for (name, _, prefix), items in dep_marks.items():
            if name in aliases:
                for item in items:
                    alias = self.matching_alias(aliases[name], item)
                    self.dep_marks.append(RelativeMark(alias, item, move_after=True))
            else:
                label = "::".join((prefix, name))
                if label in aliases:
                    for item in items:
                        alias = self.matching_alias(aliases[label], item)
                        self.dep_marks.append(
                            RelativeMark(alias, item, move_after=True)
                        )
                else:
                    sys.stdout.write(
                        f"\nWARNING: Cannot resolve the dependency marker '{name}' "
                        "- ignoring it."
                    )

    @staticmethod
    def matching_alias(aliases: List[Item], item: Item) -> Item:
        if len(aliases) == 1:
            return aliases[0]

        # handle the rare case that several tests have the same alias name
        # we use the item that best matches the node id of the dependent item
        max_matching_parts = 0
        node_id_parts = re.split("(::|/)", item.node_id)
        matching_item = aliases[0]
        for alias_item in aliases:
            alias_node_id_parts = re.split("(::|/)", alias_item.node_id)
            nr_matching_parts = 0
            for n, a in zip(node_id_parts, alias_node_id_parts):
                if n != a:
                    break
                nr_matching_parts += 1
            if nr_matching_parts > max_matching_parts:
                max_matching_parts = nr_matching_parts
                matching_item = alias_item
        return matching_item


def module_item_groups(items: List[Item]) -> Dict[str, List[Item]]:
    """
    Split items into groups per module.
    """
    module_items: OrderedDict[str, List[Item]] = OrderedDict()
    for item in items:
        module_items.setdefault(item.module_path, []).append(item)
    return module_items


def directory_item_groups(items: List[Item], level: int) -> Dict[str, List[Item]]:
    """
    Split items into groups per directory at the given level.
    The level is relative to the root directory, which is at level 0.
    """
    module_items: OrderedDict[str, List[Item]] = OrderedDict()
    for item in items:
        module_items.setdefault(item.parent_path(level), []).append(item)
    return module_items


def class_item_groups(items: List[Item]) -> Dict[str, List[Item]]:
    """
    Split items into groups per class.
    Items outside a class are sorted into a group per module.
    """
    class_items: OrderedDict[str, List[Item]] = OrderedDict()
    for item in items:
        delimiter_index = item.node_id.index("::")
        if "::" in item.node_id[delimiter_index + 2 :]:
            delimiter_index = item.node_id.index("::", delimiter_index + 2)
        class_path = item.node_id[:delimiter_index]
        class_items.setdefault(class_path, []).append(item)
    return class_items


class ScopeSorter:
    """
    Sorts the items for the defined scope.
    """

    def __init__(
        self,
        settings: Settings,
        items: List[Item],
        rel_marks: List[RelativeMark[Item]],
        dep_marks: List[RelativeMark[Item]],
        session_scope: bool = False,
    ) -> None:
        self.settings = settings
        self.items = items
        # no need to filter items in session scope
        if session_scope:
            self.rel_marks = rel_marks
            self.dep_marks = dep_marks
        else:
            self.rel_marks = filter_marks(rel_marks, self.items)
            self.dep_marks = filter_marks(dep_marks, self.items)

    def sort_items(self) -> List[Item]:
        if self.settings.group_scope.value < self.settings.scope.value:
            if self.settings.scope == Scope.SESSION:
                sorted_list = self.sort_in_session_scope()
            else:  # module scope / class group scope
                sorted_list = self.sort_in_module_scope()
        else:
            sorted_list = self.sort_items_in_scope(self.items, Scope.SESSION).items

        return sorted_list

    def sort_in_session_scope(self) -> List[Item]:
        sorted_list = []
        module_items = module_item_groups(self.items)
        if self.settings.group_scope == Scope.CLASS:
            module_groups = self.sort_class_groups(module_items)
        else:
            module_groups = [
                self.sort_items_in_scope(item, Scope.MODULE)
                for item in module_items.values()
            ]
        sorter = GroupSorter(
            Scope.MODULE, module_groups, self.rel_marks, self.dep_marks
        )
        for group in sorter.sorted_groups()[1]:
            sorted_list.extend(group.items)
        return sorted_list

    def sort_in_module_scope(self) -> List[Item]:
        sorted_list = []
        class_items = class_item_groups(self.items)
        class_groups = [
            self.sort_items_in_scope(item, Scope.CLASS) for item in class_items.values()
        ]
        sorter = GroupSorter(Scope.CLASS, class_groups, self.rel_marks, self.dep_marks)
        for group in sorter.sorted_groups()[1]:
            sorted_list.extend(group.items)
        return sorted_list

    def sort_class_groups(self, module_items: Dict[str, List[Item]]) -> List[ItemGroup]:
        module_groups = []
        for module_item in module_items.values():
            class_items = class_item_groups(module_item)
            class_groups = [
                self.sort_items_in_scope(item, Scope.CLASS)
                for item in class_items.values()
            ]
            module_group = ItemGroup()
            sorter = GroupSorter(
                Scope.CLASS, class_groups, self.rel_marks, self.dep_marks
            )
            group_order, class_groups = sorter.sorted_groups()
            module_group.extend(class_groups, group_order)
            module_groups.append(module_group)
        return module_groups

    def sort_items_in_scope(self, items: List[Item], scope: Scope) -> ItemGroup:
        item_list = ItemList(
            items, self.settings, scope, self.rel_marks, self.dep_marks
        )
        for item in items:
            item_list.collect_markers(item)

        sorted_list = item_list.sort_numbered_items()

        still_left = 0
        length = item_list.number_of_rel_groups()
        while length and still_left != length:
            still_left = length
            item_list.handle_rel_marks(sorted_list)
            item_list.handle_dep_marks(sorted_list)
            length = item_list.number_of_rel_groups()
        if length:
            item_list.print_unhandled_items()
        return ItemGroup(sorted_list, item_list.group_order())


def scope_from_name(name: str) -> Scope:
    if name == "module":
        return Scope.MODULE
    if name == "class":
        return Scope.CLASS
    return Scope.SESSION


def scoped_node_id(node_id: str, scope: Scope) -> str:
    if scope == Scope.MODULE:
        return node_id[: node_id.index("::")]
    if scope == Scope.CLASS:
        return node_id[: node_id.rindex("::")]
    return ""


class GroupSorter:
    """
    Sorts groups of items.
    """

    def __init__(
        self,
        scope: Scope,
        groups: List[ItemGroup],
        rel_marks: List[RelativeMark[Item]],
        dep_marks: List[RelativeMark[Item]],
    ) -> None:
        self.scope: Scope = scope
        self.groups: List[ItemGroup] = groups
        self.rel_marks: List[RelativeMark[ItemGroup]] = self.collect_group_marks(
            rel_marks
        )
        self.dep_marks: List[RelativeMark[ItemGroup]] = self.collect_group_marks(
            dep_marks
        )

    def collect_group_marks(
        self, marks: List[RelativeMark[Item]]
    ) -> List[RelativeMark[ItemGroup]]:
        group_marks: List[RelativeMark[ItemGroup]] = []
        for mark in marks:
            group = self.group_for_item(mark.item)
            group_to_move = self.group_for_item(mark.item_to_move)
            if group is not None and group_to_move is not None:
                group_marks.append(RelativeMark(group, group_to_move, mark.move_after))
                group_to_move.inc_rel_marks()
        return group_marks

    def group_for_item(self, item: Item) -> Optional[ItemGroup]:
        for group in self.groups:
            if item in group.items:
                return group
        return None

    def sorted_groups(self) -> Tuple[Optional[int], List[ItemGroup]]:
        group_order = self.sort_by_ordinal_markers()
        length = len(self.rel_marks) + len(self.dep_marks)
        if length == 0:
            return group_order, self.groups

        # handle relative markers the same way single items are handled
        still_left = 0
        while length and still_left != length:
            still_left = length
            self.handle_rel_marks(self.rel_marks)
            self.handle_rel_marks(self.dep_marks)
        return group_order, self.groups

    def sort_by_ordinal_markers(self) -> Optional[int]:
        start_groups = []
        middle_groups = []
        end_groups = []
        for group in self.groups:
            if group.order is None:
                middle_groups.append(group)
            elif group.order >= 0:
                start_groups.append(group)
            else:
                end_groups.append(group)
        start_groups = sorted(start_groups, key=lambda g: cast(int, g.order))
        end_groups = sorted(end_groups, key=lambda g: cast(int, g.order))
        self.groups = start_groups + middle_groups + end_groups
        if start_groups:
            group_order = start_groups[0].order
        elif end_groups:
            group_order = end_groups[-1].order
        else:
            group_order = None
        return group_order

    def handle_rel_marks(self, marks: List[RelativeMark[ItemGroup]]) -> None:
        for mark in reversed(marks):
            if move_item(mark, self.groups):
                marks.remove(mark)
